/**
 * Dasu Project
 *
 * File:         DasuManager.java
 * RCS:          $Id: DasuManager.java,v 1.23 2010/11/01 15:37:00 mas939 Exp $
 * Description:  DasuManager class (see below)
 * Author:       Mario Sanchez
 *               Northwestern Systems Research Group
 *               Department of Computer Science
 *               Northwestern University
 * Created:      May 24, 2010 at 2:27:54 PM
 * Language:     Java
 * Package:      edu.northwestern.dasu
 * Status:       Experimental (Do Not Distribute)
 *
 * (C) Copyright 2010, Northwestern University, all rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 */

package edu.northwestern.dasu;

import java.io.File;
import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;
import java.util.Arrays;
import java.util.Date;
import java.util.HashMap;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.Map;
import java.util.Properties;
import java.util.Set;
import java.util.TreeMap;
import java.util.Map.Entry;
import java.util.regex.Pattern;
import java.util.regex.Matcher;

import org.gudy.azureus2.core3.config.COConfigurationManager;
import org.gudy.azureus2.core3.config.ParameterListener;
import org.gudy.azureus2.core3.logging.ILogEventListener;
import org.gudy.azureus2.core3.logging.LogEvent;
import org.gudy.azureus2.core3.logging.LogIDs;
import org.gudy.azureus2.core3.logging.LogRelation;
import org.gudy.azureus2.core3.logging.Logger;
import org.gudy.azureus2.plugins.download.Download;
import org.gudy.azureus2.plugins.download.DownloadListener;
import org.gudy.azureus2.plugins.download.DownloadManagerListener;
import org.gudy.azureus2.plugins.utils.UTTimerEvent;

import edu.northwestern.dasu.MainGeneric;
import edu.northwestern.dasu.asn.NetworkAdminASNLookupImpl;
import edu.northwestern.dasu.drools.ActiveMeasurementEngine;
import edu.northwestern.dasu.drools.FactDidEvent;
import edu.northwestern.dasu.drools.FactMeasurementDsd;
import edu.northwestern.dasu.drools.FactMeasurementDss;
import edu.northwestern.dasu.drools.FactMeasurementLpsd;
import edu.northwestern.dasu.drools.FactMeasurementPeerRst;
import edu.northwestern.dasu.drools.FactMeasurementRst;
import edu.northwestern.dasu.drools.FactMeasurementSystem;
import edu.northwestern.dasu.drools.RuleEngine;
import edu.northwestern.dasu.measurement.*;
import edu.northwestern.dasu.stats.*;
import edu.northwestern.dasu.timer.DasuCronScheduler;
import edu.northwestern.dasu.timer.DasuMessageManager;
import edu.northwestern.dasu.timer.ITimer;
import edu.northwestern.dasu.timer.ITimerEvent;
import edu.northwestern.dasu.timer.ITimerEventPerformer;
import edu.northwestern.dasu.ui.HVPlot;
import edu.northwestern.dasu.ui.InterferenceEventUI;
import edu.northwestern.dasu.ui.UIManager;
import edu.northwestern.dasu.util.DigitalSignatureManager;
import edu.northwestern.dasu.util.Pair;
import edu.northwestern.dasu.util.PluginInterface;
import edu.northwestern.dasu.util.Util;

/**
 * @author Mario Sanchez &lt;msanchez@eecs.northwestern.edu&gt;
 * 
 *         The DasuManager class ...
 */
public class DasuManager implements Runnable {

	private static boolean isActive = false;
	public Object halt = new Object();
	private boolean terminate = false;
	public static boolean isInitialized = false;
	private int sleepInterval = DasuConfiguration.getInstance()
			.getManagerSleepInterval();
	private static long UPDATE_INTERVAL = 30 * 1000;
	public static boolean USE_SIMULATION = false;
	private static boolean DEBUG = false;
	private static DasuManager self;

	/* DROOLS ENGINES */
	public static RuleEngine engine = null;
	public static ActiveMeasurementEngine AMEngine = null;
	private static ITimer AMEngineTimer = null;
	private static ITimer LogMessageTimer = null;
	private static ILogEventListener rstListener = null;
	private static Hashtable<String, ParameterListener> COConfigurationListeners = new Hashtable<String, ParameterListener>();
	
	/* FACTS */
	public static Map<Long, FactMeasurementDsd> didFactMap = null;
	public static Map<Long, FactMeasurementDss> didDssFactMap = null;
	public static Map<Long, FactMeasurementRst> rstFactMap = null;
	public static FactMeasurementSystem systemFact = null;
	public static FactMeasurementLpsd localPeerFact = null;

	private static Map<String, Integer> dlmToDidMap = new Hashtable<String, Integer>();
	public static PersistentData persistentDataManager = new PersistentData();
	protected static PluginInterface pi = Main.getPluginInterface();
	private static Hashtable<String, Long> dldRecentState = new Hashtable<String, Long>();

	private static Hashtable<Download, DownloadListener> dldListeners =  new Hashtable<Download, DownloadListener>();
	private final HashMap<String, String> paramsEvent = new HashMap<String, String>();

	private static String persistentDataFileName = Main.getPluginInterface()
			.getPluginDirectoryName()
			+ File.separatorChar + "dasu.data." + Main.getDomainName();

	private static boolean pause = !DasuConfiguration.getInstance()
			.isRecordStats();

	private static AMCoordinatorV2 amc = null;
	private static DasuCronScheduler hcs = null;
	private static DasuMessageManager dmm = null;
	
	private static final String DASU_SERVER_LIST_LOC = DasuConfiguration.DASU_SERVER_LIST_LOC;
	private static final String DASU_DEFAULT_SERVER = DasuConfiguration.DASU_DEFAULT_SERVER;
	
	private static final boolean DASU_SKIP_SERVER_SWITCH = false;
	private static TreeMap<Double, String> DasuUrls = new TreeMap<Double, String>();
	public static String currentDasuUrl = DASU_DEFAULT_SERVER;

	public DasuManager() {
		super();
		self = this;
		isActive = true;
	}

	private static void startAMCoordinator() {
		amc = AMCoordinatorV2.getInstance();
		pi.getUtilities().createThread("DasuAMCoordinator", amc);
	}

	private static void stopAMCoordinator() {
		AMCoordinatorV2.stop();
		amc = null;
	}

	private static void startDasuCronScheduler() {
		hcs = DasuCronScheduler.getInstance();
		pi.getUtilities().createThread("DasuCronScheduler", hcs);

	}

	private static void stopDasuCronScheduler() {
		DasuCronScheduler.getInstance().stop();
		hcs = null;
	}
	
	private static void startDasuMessageManager() {
		dmm = DasuMessageManager.getInstance();
		pi.getUtilities().createThread("DasuMessageManager", dmm);
	}

	private static void stopDasuMessageManager() {
		DasuMessageManager.getInstance().stop();
		dmm = null;
	}


	private static void startRuleEngine() {

		if (isActive) {
			engine = new RuleEngine();
			engine.setDaemon(true);

			if (!engine.isAlive() && engine.isInitialized())
				engine.start();

		}
	}

	private static void startActiveEngine(String filename, Integer taskId) {

		if (AMEngine == null) {
			
			if (DEBUG)
				System.out
					.println("File Found... Starting ActiveMeasurementEngine...");
			
			AMEngine = new ActiveMeasurementEngine(filename, taskId);
			AMEngine.setDaemon(true);
			AMEngine.start();
		}

	}

	private static void launchActiveMeasrementTimer() {

		int AMECheckupInterval = DasuConfiguration.getInstance()
				.getAMECheckupInterval();
		
		if (DEBUG)
			AMECheckupInterval = 15;

		AMEngineTimer = Main.createTimer("AMEnginePeriodicTimer");
		AMEngineTimer.addPeriodicEvent(AMECheckupInterval * 1000,
				new ITimerEventPerformer() {

					public void perform(ITimerEvent event) {

						if (DEBUG)
							if (AMEngine != null)
								AMEngine.displayFacts();

						if ((AMEngine != null) && (!AMEngine.isActive())) {
							AMEngine = null;
						}

						if (AMEngine == null) {

							Pair<String, Integer> task = getAMExperiment();
							if (task != null) {
								DasuManager.startActiveEngine(task.getKey(),
										task.getValue());
							}
						}
					}

					public Pair<String, Integer> getAMExperiment() {

						String myIp = MainGeneric.getPublicIpAddress();
						String url = DasuManager.getInstance().getNewRandomDasuUrl() + DasuConfiguration.AME_SCRIPT + "?ip=" + myIp +
							"&UUIDid=" + DasuConfiguration.getInstance().getUUIDid();
						
						if ((NetworkAdminASNLookupImpl.getInstance() != null) 
								&& (NetworkAdminASNLookupImpl.getInstance().getASNInfo() != null)) {

							url = url +  "&asn=" + NetworkAdminASNLookupImpl.getInstance().getASNInfo().getAS() +
								"&cc=" + NetworkAdminASNLookupImpl.getInstance().getASNInfo().getCountry();
						}
							
						
						String interfFile = "AMExperiment.drl";
						String interfSignature = "AMExperiment.signature";
						String dir = Main.getPluginInterface()
								.getPluginDirectoryName()
								+ java.io.File.separator;

						/* load file from server */
						if (DEBUG)
							System.out.println("Connecting to " + url);
						
						Properties props = MainGeneric.downloadFromURL(url,
								5 * 1000);

						if (props != null) {

							String configUrl = null;
							Integer taskId = null;
							Integer signatureId = null;
							String signatureUrl = null;
							boolean isNewFile = false;

							for (Entry<Object, Object> ent : props.entrySet()) {
								String key = (String) ent.getKey();
								String value = (String) ent.getValue();
								if (key.equals("dasu.active.experiment")) {
									configUrl = value.split(";")[0];
									taskId = Integer
											.valueOf(value.split(";")[1]);
								} else if (key.equals("dasu.active.signature")) {
									signatureUrl = value.split(";")[0];
									signatureId = Integer.valueOf(value
											.split(";")[1]);
								}
							}

							// taskId greater than 0 are not valid active task
							// ids
							if (taskId != null
									&& signatureId != null
									&& taskId > 0
									&& taskId.equals(signatureId)
									&& !getPersistentDataManager()
											.getTaskLogHistory().containsKey(
													taskId)) {
								if (DEBUG)
									System.out
											.println("New AM experiment found... downloading");

								isNewFile = MainGeneric.downloadFileFromURL(
										configUrl, interfFile, dir);
								MainGeneric.downloadFileFromURL(signatureUrl,
										interfSignature, dir);

								if (isNewFile
										&& DigitalSignatureManager.verify(dir
												+ interfFile, dir
												+ interfSignature))
									return new Pair<String, Integer>(dir
											+ interfFile, taskId);
								else {
									if (DEBUG)
										System.out
												.println("Signature verification of experiment file failed!");
								}
							} else {
								if (DEBUG)
									System.out
											.println("AMEngine: experiment & signature taskId dont match:... "
													+ " taskId: "
													+ taskId
													+ " signatureId: "
													+ signatureId
													+ " or "
													+ " task already executed... skipping... ");
							}
						} else {
							if (DEBUG)
								System.out
										.println("AMEngine: No new experiment to load...");
						}

						return null;
					}

				});

	}

	public synchronized static DasuManager getInstance() {
		return self;
	}

	private static void loadPersistentData() {

		if (DEBUG)
			System.out.println("DasuManager(): loading persistent data");

		Object pd = MainGeneric.loadObject(persistentDataFileName);
		if (pd != null)
			persistentDataManager = (PersistentData) pd;
	}

	private static void initializeFacts() {

		didFactMap = new Hashtable<Long, FactMeasurementDsd>();
		didDssFactMap = new Hashtable<Long, FactMeasurementDss>();
		rstFactMap = new Hashtable<Long, FactMeasurementRst>();
		systemFact = new FactMeasurementSystem(new NetstatSystemStats());
		localPeerFact = new FactMeasurementLpsd(new DynamicLocalPeerStat());
		
		if (!USE_SIMULATION) {
		
			for (Download d : MainGeneric.getPluginInterface().getDownloadManager()
					.getDownloads()) {
				long did = getId(d);
				didFactMap.put(did, new FactMeasurementDsd(
						new DownloadStatsDynamic(d, 0)));
	
			}
			
			for (Download d : MainGeneric.getPluginInterface().getDownloadManager()
					.getDownloads()) {
				long did = getId(d);
				didDssFactMap.put(did, new FactMeasurementDss(
						new DownloadStatsSummary(d)));
			}
		}


	}

	private void initializeIspInfo() {

		if (persistentDataManager.getIspInfo().getKey() == null) {
			
			if ((NetworkAdminASNLookupImpl.getInstance() != null) 
				&& (NetworkAdminASNLookupImpl.getInstance().getASNInfo() != null))
			
				persistentDataManager.setIspInfo(new Pair<String, Long>(
						NetworkAdminASNLookupImpl.getInstance().getASNInfo()
								.getASName(), Util.currentGMTTime()));
		}

		if (DEBUG)
			System.out.println("!!!!!!!" + persistentDataManager.getIspInfo());

	}

	private static void launchLogMessageTimer() {

		LogMessageTimer = Main.createTimer("LogMessageTimer");
		LogMessageTimer.addPeriodicEvent(5 * 60 * 1000,
				new ITimerEventPerformer() {

					public void perform(ITimerEvent event) {
						Main.getLogger().log(
								"["
										+ new Date()
										+ "]: Monitoring "
										+ Main.getPluginInterface()
												.getDownloadManager()
												.getDownloads().length
										+ " torrents");
					}
				});

	}

	private void initialize() {

		if (!Logger.isEnabled())
			COConfigurationManager.setParameter("Logger.Enabled", true);

		loadPersistentData();
		getDasuServerList();
		refreshRSTCounter();
		refreshNetstatBWCounter();
		startAMCoordinator();
		initializeFacts();
		startRuleEngine();		
		startDasuCronScheduler();
		startDasuMessageManager();
		
		if (DasuConfiguration.getInstance().isRunExperiments())
			launchActiveMeasrementTimer();
		
		launchLogMessageTimer();
		initializeIspInfo();
		isInitialized = true;

		Main.getLogger().log(
				"[" + new Date() + "]: " + "Inference engine enabled...");

	}

	private String extractIp(String s) {

		String ip = null;
		String regex = "[\\d]{1,3}\\.[\\d]{1,3}\\.[\\d]{1,3}\\.[\\d]{1,3}:";
		Pattern p = Pattern.compile(regex);
		Matcher matcher = p.matcher(s);
		while (matcher.find()) {
			ip = matcher.group().substring(0, matcher.group().length() - 1);
		}

		return ip;

	}

	private void addDownloadListeners() {
		
		for (Download d : Main.getPluginInterface().getDownloadManager().getDownloads()) {

			if (!dldListeners.contains(d)) {
				
				DownloadListener listener = (new DownloadListener() {
					
					public void positionChanged(
							Download download, int oldPosition,
							int newPosition) {							
					}
					public void stateChanged(Download download,
							int old_state, int new_state) { 

						if ( new_state == Download.ST_STOPPED ||
								new_state == Download.ST_STOPPING ||
								new_state == Download.ST_DOWNLOADING || 
								(new_state == Download.ST_SEEDING && old_state != Download.ST_DOWNLOADING)) {
						
							if (DEBUG)
								System.out.println("addDownloadListeners(): state changed for download: " + download.getName());

							dldRecentState.put(download.getName(), System.currentTimeMillis());																		
							AMCoordinatorV2.assertToEngine(SourceType.RE, new FactDidEvent("stateChange", (long) getId(download)));
							
						}
					}
				});
				
				d.addListener(listener);
				dldListeners.put(d, listener);
						
			}
		}
	}
		
	private void removeListeners() {
			
			Iterator<Entry<Download, DownloadListener>> iter
				= dldListeners.entrySet().iterator();

			while (iter.hasNext()) {
				
				Entry<Download, DownloadListener> entry = iter.next();
				Download d = entry.getKey();
				DownloadListener listener = entry.getValue();
				d.removeListener(listener);
				iter.remove();
			}
			
			Logger.removeListener(rstListener);

			for (String param : COConfigurationListeners.keySet()) {
				COConfigurationManager.removeParameterListener(param, COConfigurationListeners.get(param));
			}
			
	}
		
	private void launchConfigurationChangeListener() {

		final HashMap<String, String> paramsEvent = new HashMap<String, String>();
		paramsEvent.put("Max Download Speed KBs", "MaxDownloadChanged");
		paramsEvent.put("Max Upload Speed KBs", "MaxUploadChanged");
		paramsEvent.put("Max.Peer.Connections.Total", "MaxConnectionsChanged");
		paramsEvent.put("Max.Peer.Connections.Per.Torrent", "MaxConnectionsPerTorrentChanged");
		paramsEvent.put("Max Seeds Per Torrent", "MaxSeedsPerTorrentChanged");
		paramsEvent.put("Use Lazy Bitfield", "LazyBitfieldChanged");
		paramsEvent.put("network.transport.encrypted.require", "EncryptionRequiredChanged");
		paramsEvent.put("network.transport.encrypted.min_level", "EncrytionLevelChanged");
		paramsEvent.put("network.transport.encrypted.fallback.outgoing", "EncryptionOutgoingChanged");
		paramsEvent.put("network.transport.encrypted.fallback.incoming", "EncryptionIncominChanged");
		paramsEvent.put("network.transport.encrypted.use.crypto.port", "EncryptionCryptoPortChanged");
		paramsEvent.put("TCP.Listen.Port", "TCPPortChanged");
		paramsEvent.put("UDP.Listen.Port", "UDPPortChanged");
		paramsEvent.put("Max.Peer.Connections.Per.Torrent.When.Seeding.Enable", "MaxConnectionsPerTorrentWhenSeedingChanged");
		paramsEvent.put("enable.seedingonly.maxuploads", "MaxUploadWhenSeedingChanged");
		
		for (String param : paramsEvent.keySet()) {
			
			ParameterListener pl = new ParameterListener() {
				public void parameterChanged(String parameterName) {
					System.out.println("CHANGED! " + parameterName);
					AMCoordinatorV2.assertToEngine(SourceType.RE, new FactDidEvent(paramsEvent.get(parameterName), -1L));
				}
			};	
			
			COConfigurationManager.addParameterListener(param, pl);
			COConfigurationListeners.put(param, pl);
			
		}		

	}
	
	private void launchDownloadListener() {

		Main.getPluginInterface().getDownloadManager().addListener(
				new DownloadManagerListener() {

					public void downloadAdded(Download download) {

						if (DEBUG) 
							System.out.println("Download added!! " + download.getName());
						
						if (!dldListeners.contains(download)) {
								
								DownloadListener listener = (new DownloadListener() {
									
									public void positionChanged(
											Download download, int oldPosition,
											int newPosition) {							
									}
									public void stateChanged(Download download,
											int old_state, int new_state) { 

										if ( new_state == Download.ST_STOPPED ||
												new_state == Download.ST_STOPPING ||
												new_state == Download.ST_DOWNLOADING || 
												(new_state == Download.ST_SEEDING && old_state != Download.ST_DOWNLOADING)) {
										
											if (DEBUG)
												System.out.println("addDownloadListeners(): state changed for download: " + download.getName());

											dldRecentState.put(download.getName(), System.currentTimeMillis());																		
											AMCoordinatorV2.assertToEngine(SourceType.RE, new FactDidEvent("stateChange", (long) getId(download)));
											
										}
									}
								});
								
								download.addListener(listener);
								dldListeners.put(download, listener);
								
								long did = getId(download);
								didFactMap.put(did, new FactMeasurementDsd(
										new DownloadStatsDynamic(download, 0)));
								
								didDssFactMap.put(did, new FactMeasurementDss(
										new DownloadStatsSummary(download)));
										
							}						
						
					}

					public void downloadRemoved(Download download) {
						
						long did = getId(download);
						synchronized(didFactMap) {
							if (didFactMap.containsKey(did))
								engine.retractFact(didFactMap.remove(did));
						}
						
						synchronized(didDssFactMap) {
							if (didDssFactMap.containsKey(did))
								engine.retractFact(didDssFactMap.remove(did));
						}

						synchronized(rstFactMap) {
							if (rstFactMap.containsKey(did))
								engine.retractFact(rstFactMap.remove(did));
						}
						
						dlmToDidMap.remove(download);
						
						synchronized(dldListeners) {
							if (dldListeners.contains(download))
								download.removeListener(dldListeners.remove(download));
						}
						
						
						if (DEBUG) 
							System.out.println("Download removed!! " + download.getName());


					}
				});
	}


	
	private void launchRSTListener() {

		rstListener = new ILogEventListener() {

			public void log(LogEvent event) {

				int type = -1;

				if (event.logID == LogIDs.PEER) {
				
					if (event.text.indexOf("Connection reset by peer") != -1)
						type = 0;
					
					else if (event.text.indexOf("An existing connection was forcibly closed by the remote host") != -1) 
						type = 1;
					
					else if (event.text.indexOf("Established") != -1) 
						type = 2;

				}
				else if ((event.logID == LogIDs.PLUGIN)) {
					return;
				}

				if (type != -1) {

					if (false)
						System.out.println("************" + event.text
								+ "********");


					for (Object obj : event.relatedTo) {
						
						if (obj instanceof LogRelation) {
							String relatedTo = ((LogRelation) obj)
									.getRelationText();
							
							//if (DEBUG) {
							//	System.out.println("obj: " + obj.getClass());
							//	System.out.println("relatedTo obj: " + relatedTo);
							//}
							
							org.gudy.azureus2.core3.peer.impl.transport.PEPeerTransportProtocol pe = 
								(org.gudy.azureus2.core3.peer.impl.transport.PEPeerTransportProtocol) obj;
							
							long timeStamp = Util.currentGMTTime();

							String tdlm = pe.getManager().getDisplayName();
							String peerip = pe.getIp();
							String bitTorrentClient = pe.getClient();
							int port = pe.getPort();
							boolean isSeed = pe.isSeed();
							boolean amSeed = pe.getManager().isSeeding();
							String source = pe.getPeerSource();
							int numSeeds = pe.getManager().getNbSeeds();
							int numPeers = pe.getManager().getNbPeers();
							String encryption = pe.getEncryption();
							
							//boolean isCryptoRequired = com.aelitis.azureus.core.networkmanager.NetworkManager.getCryptoRequired(pe.getManager().getAdapter().getCryptoLevel());
							
							if (!dlmToDidMap.containsKey(tdlm))
								return;
							
							long did = dlmToDidMap.get(tdlm);

							if (false) {
								System.out.println("tdlm: " + tdlm);
								System.out.println("peerip: " + peerip);
								System.out.println("client: " + bitTorrentClient);
								System.out.println("port: " + port);
								System.out.println("isSeed: " + isSeed);
								System.out.println("amSeed?: " + amSeed);
								System.out.println("source: " + source);
								System.out.println("numSeeds: " + numSeeds);
								System.out.println("numPeers: " + numPeers);
								System.out.println("encryption: " + encryption);
							}

							Main.getLogger().log("[" + new Date() + "]: " + event.text + " (" + peerip + ")");
							
							if ( type == 0 || type ==1 ) {
							
								RSTStat rstStat = new RSTStat(did, peerip, timeStamp, 1, type, 
										port, bitTorrentClient, source, isSeed, amSeed,
										numSeeds, numPeers, encryption);
	
								/*if the download changed status recently ignore reset*/							
								if (DasuConfiguration.getInstance().getIgnoreStateChangeRST()) {
	
									if (dldRecentState.containsKey(tdlm)) {
										
										if (System.currentTimeMillis() - dldRecentState.get(tdlm) < DasuConfiguration.getInstance().getStateChangeRSTInterval()*1000 ) {
											if (DEBUG) {
												System.out.println("****Ignore reset... due to download state change****");
											}
										
											return;
										}
									}
								}
								
							
								Statistics.getInstance().addRSTStatResult(
										rstStat);
								
								if (!rstFactMap.containsKey(did))
									rstFactMap.put(did, new FactMeasurementRst(rstStat));
								else
									rstFactMap.get(did).addEntry(
											rstStat);		
								
								if (engine != null) {
									engine.addFact(rstFactMap.get(did));
									engine.addFact(new FactMeasurementPeerRst(rstStat));
								}
							}
//							else if (type == 2) {
//								
//								RemotePeerConnectionStat rpcStat = new RemotePeerConnectionStat(peerip, timeStamp, source, 
//										port, bitTorrentClient, isSeed, amSeed, numSeeds, numPeers, encryption);
//								
//							
//								Statistics.getInstance().addRpcStatResult(
//										rpcStat);
//								
//							}

						}
					}
				}
			}
		};

		Logger.addListener(rstListener);

	}
	
	
	private static void refreshRSTCounter() {

		if (edu.northwestern.dasu.Main.perf != null) {
			edu.northwestern.dasu.Main.perf.perform(new UTTimerEvent() {
				public void cancel() {
				}
			});
		}
	}
	
	private static void refreshNetstatBWCounter() {

		if (edu.northwestern.dasu.Main.netstatBWPerf != null) {
			edu.northwestern.dasu.Main.netstatBWPerf.perform(new UTTimerEvent() {
				public void cancel() {
				}
			});
		}
		
	}
	
	private void printDebug() {
	
		if ((localPeerFact != null) && (systemFact != null)) {

			System.out.println("\n" + "local lpsd: avgUploadRate: "
					+ (localPeerFact.getBytesSent()/(localPeerFact.getShortDeltaTime()!=0?localPeerFact.getShortDeltaTime():1)/1000) + " kB/s"
					+ " avgDownloadRate: "
					+ (localPeerFact.getBytesReceived()/(localPeerFact.getShortDeltaTime()!=0?localPeerFact.getShortDeltaTime():1)/1000) + " kB/s"
					+ " deltaTime: "
					+ (localPeerFact.getShortDeltaTime()));
			
			System.out.println("local netstat: avgUploadRate: "
					+ (systemFact.getBytesSent()/(systemFact.getShortDeltaTime()!=0?systemFact.getShortDeltaTime():1)/1000) + " kB/s"
					+ " avgDownloadRate: "
					+ (systemFact.getBytesReceived()/(systemFact.getShortDeltaTime()!=0?systemFact.getShortDeltaTime():1)/1000) + " kB/s"
					+ " deltaTime: "
					+ (systemFact.getShortDeltaTime()));

			System.out.println("local lpsd: uploadRate: "
					+ localPeerFact.getUploadRate()
					+ " avgUpload: "
					+ localPeerFact.getAverageUpload()
					+ " stdevUpload: "
					+ localPeerFact.getStdevUpload()
					+ " downloadRate: "
					+ localPeerFact.getDownloadRate()
					+ " avgDownload: "
					+ localPeerFact.getAverageDownload()
					+ " stdevDownload: "
					+ localPeerFact.getStdevDownload()
					+ " deltaTime: "
					+ localPeerFact.getLongDeltaTime());
			
			System.out.println("local netstat: bytesSent: "
					+ systemFact.getBytesSent()
					+ " avgBytesSent: "
					+ systemFact.getAverageBytesSent()
					+ " stdevBytesSent: "
					+ systemFact.getStdevBytesSent()
					+ " bytesReceived: "
					+ systemFact.getBytesReceived()
					+ " avgBytesReceived: "
					+ systemFact.getAverageBytesReceived()
					+ " stdevBytesReceived: "
					+ systemFact.getStdevBytesReceived()
					+ " deltaTime: "
					+ systemFact.getLongDeltaTime());
			
			System.out.println("local lpsd: numTorrents: "
					+ localPeerFact.getNumTorrents()
					+ " numConnctedSeeds: "
					+ localPeerFact.getNumConnectedSeeds()
					+ " avgNumConnectedSeeds: "
					+ localPeerFact.getAverageNumConnectedSeeds()
					+ " numConnctedLeechers: "
					+ localPeerFact.getNumConnectedLeechers()
					+ " avgNumConnectedLeechers: "
					+ localPeerFact.getAverageNumConnectedLeechers()
					+ " maxPeerConnections: "
					+ localPeerFact.getMaxPeerConnectionsGlobal()
					+ " maxPeerConnectionsPerTorrent: "
					+ localPeerFact.getMaxPeerConnectionsPerTorrent()
					+ " maxPeerConnectionsPerTorrentSeeding: "
					+ localPeerFact.getMaxPeerConnectionsPerTorrentSeeding()
					+ " maxUploadRateSeeding: "
					+ localPeerFact.getMaxUploadRateSeeding()
					+ " maxSeedsPerTorrent: "
					+ localPeerFact.getMaxSeedsPerTorrent()
					+ " lazyBitfield: "
					+ localPeerFact.isLazyBitfield()
					+ " requireEncryption: "
					+ localPeerFact.isRequireEncryption()
					+ " encryptionMinLevel: "
					+ localPeerFact.getEncryptionMinLevel()
					+ " allowNonEncryptedOut: "
					+ localPeerFact.isAllowNonEncryptedOut()
					+ " allowNonEncryptedIn: "
					+ localPeerFact.isAllowNonEncryptedIn()
					+ " useCryptoPortExtension: "
					+ localPeerFact.isUseCryptoPortExtension()
					+ " tcpPortIn: " 
					+ localPeerFact.getTcpPortIn()
					+ " udpPortIn: "
					+ localPeerFact.getUdpPortIn());
			
			System.out.println();
		}
		
		Set<Long> dids = didFactMap.keySet();
		for (long did : dids) {

//		for (Download d : MainGeneric.getPluginInterface()
//				.getDownloadManager().getDownloads()) {
//			long did = getId(d);

			synchronized(didFactMap) {
				
				if (didFactMap.containsKey(did)) {
				   System.out
						.println("Did: " + did + " uploadRate: "
								+ didFactMap.get(did).getUploadRate()
								+ " avgUpload: "
								+ didFactMap.get(did)
										.getAverageUpload()
								+ " stdevUpload: "
								+ didFactMap.get(did).getStdevUpload()
								+ " downloadRate: "
								+ didFactMap.get(did).getDownloadRate()
								+ " avgDownload: "
								+ didFactMap.get(did)
										.getAverageDownload()
								+ " stdevDownload: "
								+ didFactMap.get(did)
										.getStdevDownload() + " did: "
								+ didFactMap.get(did).getDownloadId()
								+ " completed: "
								+ didFactMap.get(did).getCompleted()
								+ " numRst: "
								+ didFactMap.get(did).getNumRst());
				   
				   System.out.println("Did: " + did + " numPeers: " 
						   + didFactMap.get(did).getNumPeers()
						   + " avgNumPeers: " 
						   + didFactMap.get(did).getAverageNumPeers()
						   + " connectedSeeds: "
						   + didFactMap.get(did).getConnectedSeeds()
						   + " avgConnectedSeeds: "
						   + didFactMap.get(did).getAverageConnectedSeeds()
						   + " connectedLeechers: "
						   + didFactMap.get(did).getConnectedLeechers()
						   + " avgConnectedLeechers: "
						   + didFactMap.get(did).getAverageConnectedLeechers()									   
						   + " numPendingPeers: " 
						   + didFactMap.get(did).getNumPendingPeers());
				   
				}
			} //synchronized
			
		}

		if (engine != null)
			engine.displayFacts();
		
	}
	
	
	public void testing() {
	
		if (DEBUG) {
			System.out.println("Disabling AME experiments due to DEBUG mode...");
			DasuConfiguration.getInstance().setRunExperiments(false);
		}
		
	//    	if (!AMCoordinatorV2.historicSampleExists("time_of_day_bw_limit_day_minute_5")) {
	//		AMCoordinatorV2.historicSampleCreate("time_of_day_bw_limit_day_minute_5", TimeType.DAY, TimeType.MINUTE, 5);
	//
	//    ArrayList maxDownloadRateLpsdPastTwoHours = (ArrayList) AMCoordinatorV2.historicSampleGetValuesSubsetPast(
	//            "time_of_day_bw_limit_day_minute_5", "maxDownloadRateLpsd", Util.currentDasuTime(), 60);
	//        System.out.println("\n\n\n\n\n\nENGINE: " + maxDownloadRateLpsdPastTwoHours);
		
	//	SignalSample ss = null;
	//	try {
	//		ss = new SignalSample("uploadRate", TimeType.DAY, TimeType.HOUR, 1, "BytesPerSecond");
	//	} catch (Exception e1) {
	//		e1.printStackTrace();
	//	}
	//	System.out.println(new Date(Util.currentDasuTime()));
	//	ss.set(Util.currentDasuTime(), 4532L);
	//	
	//	System.out.println(ss.getName());
	//	System.out.println(ss.getValuesHash());
	//	System.out.println(ss.getTimeStampsHash());
		
	//	Collection<Float> nt = AMCoordinatorV2.historicSampleGetValues("netstat_max_downloadRate_hour_minute_1");
	//	Collection<Float> bt  = AMCoordinatorV2.historicSampleGetValues("bt_max_downloadRate_hour_minute_1");
	//	System.out.println(SignalManipulation.evalAverage(new Float[]{1f, 2f, 3f}, 1));
		
	//	System.out.println(nt);
	//	System.out.println(bt);
	//	

		Float[] nt = new Float[]{77.0f, 65.0f, 55.0f, 58.0f, 52.0f, 52.0f, 76.0f, 61.0f, 58.0f, 59.0f, 
				67.0f, 62.0f, 67.0f, 76.0f, 77.0f, 65.0f, 55.0f, 58.0f, 51.0f, 52.0f, 73.0f, 66.0f, 78.0f, 57.0f, 
				67.0f, 72.0f, 52.0f, 76.0f, 58.0f, 59.0f, 67.0f, 62.0f, 67.0f, 76.0f, 77.0f, 65.0f, 55.0f, 
				58.0f, 51.0f, 52.0f, 73.0f, 66.0f, 78.0f, 57.0f, 67.0f, 72.0f, 52.0f, 76.0f, 61.0f, 58.0f, 59.0f, 
				67.0f, 62.0f, 67.0f, 76f, 73.0f, 66.0f, 78.0f, 57.0f};
		
		Float[] bt = new Float[]{
				40.0f, 40.0f, 43.0f, 41.0f, 44.0f, 50.0f, 55.0f, 42.0f, 12.0f, 4.0f, 10.0f, 12.0f, 11.0f, 14.0f, 
				44.0f, 50.0f, 53.0f, 48.0f, 40.0f, 53.0f, 41.0f, 44.0f, 42.0f, 50.0f, 55.0f, 42.0f, 59.0f,
				44.0f, 50.0f, 66.0f, 51.0f, 40.0f, 43.0f, 41.0f, 44.0f, 42.0f, 50.0f, 55.0f, 42.0f, 59.0f, 44.0f, 50.0f, 
				66.0f, 59.0f, 66.0f, 44.0f, 50.0f, 53.0f, 48.0f, 40.0f, 66.0f, 44.0f, 50.0f, 53.0f, 48.0f, 53.0f,
				42.0f, 50.0f, 55.0f};
	
		Float[] signal = new Float[]{0f, 0f, 0f, 0f, 0f, 0f, 0f, 0f, 0f, 0f, 0f, 0f, 0f, 0f, 0f, 0f, 0f, 0f, 0f, .0f, 24.0f, 34f, 13f, 44f, 53f}; 
	
		Long[] time = new Long[bt.length];
		for (int i=0; i < bt.length; i++ )
			time[i] = Util.currentDasuTime()+i*1000*60;
	
		System.out.println("signal: " + Arrays.asList(signal));
		System.out.println("getMean: " + SignalManipulation.getAverage(signal));
		System.out.println("getStdev: " + SignalManipulation.getStdDev(signal));
		System.out.println("getPercentile: " + SignalManipulation.getPercentile(signal, 95));
		System.out.println("getPercentileAboveValue: " + SignalManipulation.getPercentileAboveValue(signal, 95, SignalManipulation.getAverage(signal)));
		System.out.println("getPercentileIndexPosition: " + SignalManipulation.getPercentileIndexPosition(signal, 95));
		
		MovingWindow mw =  new MovingWindow(bt, time, 5);
		System.out.println(mw);	
		
		InterferenceEventUI eventUI = new InterferenceEventUI("test event");
	    eventUI.setStatus("Detected");
	    eventUI.addFullScreenDescriptionLine("Netstat Plot: just to see how it looks");
		eventUI.addDescriptionLine(new String[]{"Mean"});
		eventUI.addPlot("mean", mw.getTime(), mw.getMean(), "(time)", "(kbps)", HVPlot.LINE_GRAPH, HVPlot.FORMAT_TIME);
		eventUI.addDescriptionLine(new String[]{"signal"});
		eventUI.addPlot("variance from stdev", null, mw.getSignal(), "(time)", "(kbps)", HVPlot.SCATTER_GRAPH, HVPlot.FORMAT_DAY_WEEK);
		eventUI.addDescriptionLine(new String[]{"variance from stdev"});
		eventUI.addPlot("variance from stdev", null, mw.getVarianceFromStdev(), "(time)", "(kbps)", HVPlot.LINE_GRAPH, HVPlot.FORMAT_MONTH);
	    AMCoordinatorV2.sendToLog("Dasu plot test");
		AMCoordinatorV2.getUIManager().setEvent("NetstatPlot", eventUI);
		
		System.out.println("\n");
		System.out.println("correlation 5: " + SignalManipulation.getCorrelationCollection(nt, bt, 5));
		System.out.println("correlation 5 cumulative: " + SignalManipulation.getCorrelationCollection(
				SignalManipulation.getCumulative(nt), SignalManipulation.getCumulative(bt), 5));
		
		System.out.println("time: (" + time.length + ") " + Arrays.asList(time));
		System.out.println("cumulative bt: (" + SignalManipulation.getCumulative(bt).size() + ") " + SignalManipulation.getCumulative(bt));
	
		System.out.println("eval Slope 1: " + SignalManipulation.getSlopeCollection( 
				SignalManipulation.getCumulative(bt), Arrays.asList(time), 2));
		
	}
	
	
	public  String getHexString(byte[] b) {
		
		String result = "";
		for (int i=0; i < b.length; i++) {
			result +=
		          Integer.toString( ( b[i] & 0xff ) + 0x100, 16).substring( 1 );
		}
		return result;
	}
	
	private static byte[] hexStringToByteArray(String s) {
		
		byte[] b = new byte[s.length() / 2];
	    for (int i = 0; i < b.length; i++){
	    	int index = i * 2;
	    	int v = Integer.parseInt(s.substring(index, index + 2), 16);
	    	b[i] = (byte)v;
	    }
	    return b;
	}
	
	/**
	 * Perform periodic data-collections
	 */
	
	public void run() {
		
		initialize();	

		//testing();

		if (!USE_SIMULATION) {
			launchRSTListener();
			launchDownloadListener();
			launchConfigurationChangeListener();
			
		} else {
			localPeerFact = null;
			systemFact = null;
		}
		
		while (isActive) {

			if (USE_SIMULATION) {
				performSimulation();
			} else {
				
				/*
				 * update measurement facts associated with the collected
				 * variables
				 */
				
				refreshRSTCounter();
				refreshNetstatBWCounter();

				if (engine != null) {

					//Rpss facts are inserted by the PeerStats Manager 
					//asynchronously as peers disconnect, only for peers
					//that actually transmitted/received something from us
					
					for (Download d : MainGeneric.getPluginInterface()
							.getDownloadManager().getDownloads()) {
						long did = getId(d);

						//refreshRSTCounter(); /*why again?*/
						
						int rst = 0;
						if (rstFactMap != null)
							rst = (rstFactMap.containsKey(did)) ? rstFactMap
									.get(did).getNumRst() : 0;

						synchronized(didFactMap) {

							if (!didFactMap.containsKey(did)) {
								didFactMap.put(did, new FactMeasurementDsd(
										new DownloadStatsDynamic(d, rst)));
							} else {
								didFactMap.get(did).addEntry(
										new DownloadStatsDynamic(d, rst));
							}
	
							engine.addFact(didFactMap.get(did));
						}
						
						synchronized(didDssFactMap) {

							if (!didDssFactMap.containsKey(did)) {
								didDssFactMap.put(did, new FactMeasurementDss(
										new DownloadStatsSummary(d)));
							} else {
								didDssFactMap.get(did).addEntry(
										new DownloadStatsSummary(d));
							}
	
							engine.addFact(didDssFactMap.get(did));
						}

					}
					
					systemFact.addEntry(new NetstatSystemStats());
					localPeerFact.addEntry(new DynamicLocalPeerStat());

					engine.addFact(systemFact);
					engine.addFact(localPeerFact);

				}

				Iterator<Entry<Long, FactMeasurementRst>> iter = rstFactMap
						.entrySet().iterator();
				while (iter.hasNext()) {
					Entry<Long, FactMeasurementRst> e = iter.next();
					e.getValue().updateEntryList();
				}

				if (DEBUG)
					printDebug();

			}// else

			if (DEBUG)
				System.out.println("DasuManager: sleeping for " + sleepInterval
						+ " seconds...");

			synchronized (halt) {

				try {
					if (!terminate) {
						if (USE_SIMULATION)
							Thread.sleep(2);
						else
							halt.wait(sleepInterval * 1000);
					}
				} catch (InterruptedException e) {
					// TODO Auto-generated catch block
					e.printStackTrace();
				}

			}

			if (terminate) {
				self = null;
				return;
			}

		}// while

	}

	private boolean performSimulation() {

		if (!InterferenceSimulator.isActive()) {
			if (!InterferenceSimulator.loadTimeline())
				return false;
			InterferenceSimulator.printEvents();
		}

		
		//why do you need this?
		if (edu.northwestern.dasu.Main.perf != null) {
			edu.northwestern.dasu.Main.perf.perform(new UTTimerEvent() {
				public void cancel() {
				}
			});
		}

		/*
		 * collect all sample variables first to minimize time discrepancy
		 * between samples
		 */

		HashMap<Long, DownloadStatsDynamic> didTempMap = InterferenceSimulator
				.getModifiedDownloadStatsDynamicMap();

		HashMap<Long, RSTStat> didRstMap = InterferenceSimulator
				.getModifiedRSTStatMap();

		NetstatSystemStats systemRst = InterferenceSimulator.getModifiedNetstatSystemStats();
		
		DynamicLocalPeerStat lpsd = InterferenceSimulator.getModifiedDynamicLocalPeerStat();

		for (Entry<Long, RSTStat> e : didRstMap.entrySet()) {

			System.out.println("performSimulation(): did: "
					+ e.getValue().downloadId + " numRst: "
					+ e.getValue().numRst + " time: "
					+ e.getValue().timeRecorded);
		}

		/*
		 * update/create measurement facts associated with the collected
		 * variables
		 */

		if (systemFact == null) {
			systemFact = new FactMeasurementSystem(systemRst);
			if (engine != null)
				engine.addFact(systemFact);
		} else
			systemFact.addEntry(systemRst);

		if (localPeerFact == null) {
			localPeerFact = new FactMeasurementLpsd(lpsd);
			if (engine != null)
				engine.addFact(localPeerFact);
		} else
			localPeerFact.addEntry(lpsd);

		for (long did : didTempMap.keySet()) {

			if (!didFactMap.containsKey(did)) {

				didFactMap
						.put(did, new FactMeasurementDsd(didTempMap.get(did)));
				if (engine != null)
					engine.addFact(didFactMap.get(did));
			} else
				didFactMap.get(did).addEntry(didTempMap.get(did));
		}

		for (long did : didRstMap.keySet()) {

			if (!rstFactMap.containsKey(did))
				rstFactMap.put(did, new FactMeasurementRst(didRstMap.get(did)));
			else
				rstFactMap.get(did).addEntry(didRstMap.get(did));

			if (engine != null) {
				engine.addFact(rstFactMap.get(did));
				//System.out.println("\n\n\n\nENGINE: " + engine + "\n\n\n\n");
			}

			// engine.updateFact(rstFactMap.get(did));

			// System.out.println(rstFactMap);

		}

		Iterator<Entry<Long, FactMeasurementRst>> iter = rstFactMap.entrySet()
				.iterator();
		while (iter.hasNext()) {
			Entry<Long, FactMeasurementRst> e = iter.next();
			e.getValue().updateEntryList();
			// if (e.getValue().updateEntryList() == 0) {
			// engine.retractFact(e.getValue());
			// iter.remove();
			// }
		}

		if (DEBUG)
			printDebug();

		InterferenceSimulator.increaseTick();

		return true;

	}

	private static int getId(Download d) {

		MessageDigest md;
		try {
			md = MessageDigest.getInstance("SHA");

			md.update(d.getName().getBytes());
			md.update(Util.convertLong(d.getCreationTime()));
			byte hash[] = md.digest();
			String s = new String(hash);
			dlmToDidMap.put(d.getName(), s.hashCode());
			return s.hashCode();

		} catch (NoSuchAlgorithmException e) {
			// TODO Auto-generated catch block
			// e.printStackTrace();
		}

		dlmToDidMap.put(d.getName(), (int) (d.getName().hashCode() + d
				.getCreationTime()));
		return (int) (d.getName().hashCode() + d.getCreationTime());

	}

	public static void setRecordingInterval(int interval) {
		UPDATE_INTERVAL = interval;

	}

	private static void stopActiveMeasrementTimer() {
		if (AMEngineTimer != null)
			AMEngineTimer.destroy();
	}

	private static void stopLogMessageTimer() {
		if (LogMessageTimer != null)
			LogMessageTimer.destroy();
	}

	public void stop() {

		isActive = false;
		synchronized (halt) {
			terminate = true;
		}

		MainGeneric.saveObject(persistentDataManager, persistentDataFileName);
		
		stopLogMessageTimer();
		stopActiveMeasrementTimer();
		stopDasuCronScheduler();
		stopDasuMessageManager();

		if (DEBUG)
			System.out.println("Stopping AMCoordinator");

		if (AMCoordinatorV2.isActive())
			stopAMCoordinator();

		if (DEBUG)
			System.out.println("Stopping rule engines");

		stopEngines();
		removeListeners();

		Main.getLogger().log(
				"[" + new Date() + "]: " + "Inference engine disabled...");
		try {
			finalize();
		} catch (Throwable e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}

		if (DEBUG)
			System.out.println("Done...");

	}

	private static void stopFacts() {

		didFactMap = null;
		didDssFactMap = null;
		rstFactMap = null;
		systemFact = null;
		localPeerFact = null;

	}

	private static void stopEngines() {

		if (engine != null) {
			if (engine.isAlive())
				engine.stopEngine();
		}

		if (AMEngine != null) {
			if (AMEngine.isAlive())
				AMEngine.stopEngine();
		}

	}

	public static PersistentData getPersistentDataManager() {
		return persistentDataManager;
	}

	public synchronized static UIManager getUIManager() {
		return persistentDataManager.getUiManager();
	}

	public static boolean isPause() {
		return pause;
	}

	public void setPause(boolean pause) {

		stop();

	}
	
	/**
	 * 
	 */
	public void getDasuServerList() {
		
		DasuUrls.put(1.0, DASU_DEFAULT_SERVER);
		Properties servers = MainGeneric.downloadFromURL(DASU_SERVER_LIST_LOC,
				30 * 1000);
		for (Entry<Object, Object> ent : servers.entrySet()) {
			try {
				String pair[] = ((String) ent.getValue()).split(";");
				DasuUrls.put(Double.parseDouble(pair[1]), pair[0]);
			} catch (Exception e) {
				e.printStackTrace();
			}

		}
		// remove default if there's more than one server and its weight wasn't
		// changed
		if (DASU_SKIP_SERVER_SWITCH) 
			return;
		
		if (DasuUrls.size() > 1 && DasuUrls.get(1.0).equals(DASU_DEFAULT_SERVER))
			DasuUrls.remove(1.0);

		double val = MainGeneric.getRandom().nextDouble();
		for (Entry<Double, String> ent : DasuUrls.entrySet()) {
			if (val <= ent.getKey()) {
				currentDasuUrl = ent.getValue();
				break;
			}
		}
	}
	
	public String getNewRandomDasuUrl() {

		String thisUrl = currentDasuUrl;
		String oldUrl = currentDasuUrl;
		
		if (DasuUrls.size() <= 1) {
			currentDasuUrl = thisUrl;
			return thisUrl;
		}
		else {
			do {
				double val = MainGeneric.getRandom().nextDouble();
				for (Entry<Double, String> ent : DasuUrls.entrySet()) {
					if (val <= ent.getKey())
						thisUrl = ent.getValue();
				}
			} while (!thisUrl.equals(oldUrl));
		}
		currentDasuUrl = thisUrl;
		return thisUrl;
	}

}
