/**
 * Dasu Project
 *
 * File:         RuleEngine.java
 * RCS:          $Id: RuleEngine.java,v 1.8 2010/11/01 15:37:00 mas939 Exp $
 * Description:  RuleEngine class (see below)
 * Author:       Mario Sanchez
 *               Northwestern Systems Research Group
 *               Department of Computer Science
 *               Northwestern University
 * Created:      May 31, 2010 at 5:28:22 PM
 * Language:     Java
 * Package:      edu.northwestern.dasu.drools
 * Status:       Experimental (Do Not Distribute)
 *
 * (C) Copyright 2010, Northwestern University, all rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 */
package edu.northwestern.dasu.drools;

import java.io.DataInputStream;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.net.URL;
import java.net.URLClassLoader;
import java.util.Iterator;
import java.util.Properties;
import java.util.Map.Entry;

import org.drools.event.rule.ActivationCancelledEvent;
import org.drools.event.rule.ActivationCreatedEvent;
import org.drools.event.rule.AfterActivationFiredEvent;
import org.drools.event.rule.AgendaEventListener;
import org.drools.event.rule.AgendaGroupPoppedEvent;
import org.drools.event.rule.AgendaGroupPushedEvent;
import org.drools.event.rule.BeforeActivationFiredEvent;
import org.drools.logger.KnowledgeRuntimeLoggerFactory;
import org.drools.runtime.rule.FactHandle;

import edu.northwestern.dasu.DasuConfiguration;
import edu.northwestern.dasu.DasuManager;
import edu.northwestern.dasu.Main;
import edu.northwestern.dasu.MainGeneric;
import edu.northwestern.dasu.measurement.AMCoordinatorV2;
import edu.northwestern.dasu.util.DigitalSignatureManager;


/**
 * @author Mario Sanchez &lt;msanchez@eecs.northwestern.edu&gt;
 *
 * The RuleEngine class ...
 */
public class RuleEngine extends RuleBasedEngine {
	
	public RuleEngine() {
		super("RuleEngine");
	}
	
	@Override
	protected void loadFacts() {
	
		/* facts addition */
		addFact(DasuManager.systemFact);
		addFact(DasuManager.localPeerFact);
		Iterator<FactMeasurementDsd> iter = DasuManager.didFactMap.values().iterator();
		while (iter.hasNext())
			addFact(iter.next());
		
		Iterator<FactMeasurementDss> iter2 = DasuManager.didDssFactMap.values().iterator();
		while (iter2.hasNext())
			addFact(iter2.next());
		
	}
	
	private boolean extractFromjar(String fileToExtract) throws IOException {
				
		boolean ret = false;
		ClassLoader classLoader = Main.getPluginInterface().getPluginClassLoader();

		if (classLoader instanceof URLClassLoader) {
			URLClassLoader current = (URLClassLoader) classLoader;

			URL url = current.findResource(fileToExtract);
			if (url != null) {
	        
			DataInputStream in = new DataInputStream(url.openStream());	        
		    byte[] bytes = new byte[in.available()];
		  
		    int offset = 0;
		    int numRead = 0;
		    while (offset < bytes.length
		             && (numRead=in.read(bytes, offset, bytes.length-offset)) >= 0) {
		          offset += numRead;
		    }
	        
	        String filename = Main.getPluginInterface().getPluginDirectoryName()
			+ java.io.File.separator + fileToExtract;
	        
	        FileOutputStream sigfos = new FileOutputStream(filename);
	        sigfos.write(bytes);
	        sigfos.close();
			ret = true;
	
			}
		}
		
		return ret;
	}
		
	@Override
	public boolean loadRules() {
		boolean verified = false;

		String url = DasuManager.getInstance().getNewRandomDasuUrl();
		url = url + DasuConfiguration.CONFIG_LIST;
		
		String interfFile = "interferencedetectionRules.drl";
        String interfSignature = "interferencedetectionRules.signature";
        String dir  = Main.getPluginInterface().getPluginDirectoryName()       
			+ java.io.File.separator;
        boolean isNewFile = false;
        
		/*load file from server*/
        Properties props = MainGeneric.downloadFromURL(url,  2*1000);
//		if (DEBUG) {
//			System.out.println("RuleEngine: skipping download due to DEBUG mode");
//			props=null;
//		}

		if (props != null) {

			String configUrl = null;
			Float configVersion = null;
			String signatureUrl = null;
			
			for (Entry<Object, Object> ent : props.entrySet()){				
				String key = (String)ent.getKey();
				String value = (String)ent.getValue();
				if (key.equals("dasu.inference.config")){
					configUrl = value.split(";")[0];
					configVersion = Float.valueOf(value.split(";")[1]);
				} else if (key.equals("dasu.inference.signature")) {
					signatureUrl = value.split(";")[0];
				}	
			}

			/* 
			 * if there's a new version of interference files put them in the plugin directory 
			 * else nothing new from the server
			 */
			
			if ((configVersion != null) && 
					((configVersion > DasuConfiguration.getInstance().getInferenceVersion()) ||
							!(new File(dir+interfFile).isFile()))) {
				
				if (DEBUG) {
					System.out.println("RuleEngine: Inference file current version: " + DasuConfiguration.getInstance().getInferenceVersion()
						+ " server version: " + configVersion);
					System.out.println("Downloading file " + configUrl);
				}
				
				isNewFile = MainGeneric.downloadFileFromURL(configUrl, interfFile, dir);
				MainGeneric.downloadFileFromURL(signatureUrl, interfSignature, dir);
			}
			
			if (isNewFile) {
				
				if (DEBUG)
					System.out.println("RuleEngine: Verify signature of downloaded file...");
				
				verified = DigitalSignatureManager.verify(dir+interfFile, dir+interfSignature);
				if (verified) {
					DasuConfiguration.getInstance().setInferenceVersion(configVersion);
			        MainGeneric.updateProperties();
			        AMCoordinatorV2.sendToLog("New interference definition file downloaded (" + configVersion + ")" );

				}
			}
		}
		
		/*
		 * If no new version from server try to load existing file 
		 * located in pluginDirectory (if any)
		 */
		if ( !verified &&  (new File(dir+interfFile).isFile()) 
				&& (new File(dir+interfSignature).isFile()) ) {
			
			if (DEBUG)
				System.out.println("RuleEngine: File already exists!! Verify signature of exising file...");

			verified = DigitalSignatureManager.verify(dir+interfFile, dir+interfSignature);
		}
		
		/* else use file from jar */
		if (!verified) {

			try {
				if ( !extractFromjar(interfFile) || !extractFromjar(interfSignature) )
					return false;
				
				if (DEBUG)
					System.out.println("RuleEngine: No inference file found... loading from jar..." +
						" and verify signature of exising file...");
				
				verified = DigitalSignatureManager.verify(dir+interfFile, dir+interfSignature);
				System.out.println(dir+interfFile + " " +  dir+interfSignature + " " + verified);

			} catch (IOException e) {

				try {
					throw (new Exception("No Interferece Detection Rules file from jars"));
				} catch (Exception e1) {
					e1.printStackTrace();
				}
										
			}
		}
		
		//SKIP SIGNATURE VERIFICATION
		//verified = true;
		
		if (verified) {
			kbase = loadRules(dir+interfFile);
			ksession = kbase.newStatefulKnowledgeSession();

			ksession.addEventListener(new AgendaEventListener() {

				public void activationCancelled(ActivationCancelledEvent event) {			
		
				}
		
				public void activationCreated(ActivationCreatedEvent event) {
		
				}
		
				public void afterActivationFired(AfterActivationFiredEvent event) {

					for (FactHandle fh : event.getActivation().getFactHandles()) {
						//System.out.println("!!!!!!!!afterActivationFired: " + fh.toExternalForm());
						getModifiedFacts().remove(fh);
					}
				}
		
				public void agendaGroupPopped(AgendaGroupPoppedEvent event) {
		
				}
		
				public void agendaGroupPushed(AgendaGroupPushedEvent event) {
		
				}
		
				public void beforeActivationFired(BeforeActivationFiredEvent event) {
		
					for (FactHandle fh : event.getActivation().getFactHandles()) {
							//System.out.println("!!!!!!!!beforeActivationFired: " + fh.toExternalForm());
							getModifiedFacts().add(fh);
					}
		
				}
				
			});
			
			if (DEBUG)
				logger = KnowledgeRuntimeLoggerFactory.newFileLogger(ksession, "test");
		}
		else
			if (DEBUG)
				System.out.println("Verification of all Inference files failed!");
		
		return verified;

	}
	
	
}
