/**
 * Dasu Project
 *
 * File:         Statistics.java
 * RCS:          $Id: Statistics.java,v 1.9 2010/11/01 15:37:00 mas939 Exp $
 * Description:  Statistics class (see below)
 * Author:       David Choffnes
 *               Northwestern Systems Research Group
 *               Department of Computer Science
 *               Northwestern University
 * Created:      Jun 14, 2006 at 7:51:32 AM
 * Language:     Java
 * Package:      edu.northwestern.news.stats
 * Status:       Experimental (Do Not Distribute)
 *
 * (C) Copyright 2006, Northwestern University, all rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 */
package edu.northwestern.dasu.stats;

import java.io.IOException;
import java.sql.SQLException;
import java.util.Date;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedHashSet;
import java.util.LinkedList;
import java.util.Map;
import java.util.Random;
import java.util.Set;
import java.util.Map.Entry;

import edu.northwestern.dasu.DasuConfiguration;
import edu.northwestern.dasu.Main;
import edu.northwestern.dasu.MainGeneric;
import edu.northwestern.dasu.database.DatabaseReporter;
import edu.northwestern.dasu.database.PhpReporter;
import edu.northwestern.dasu.drools.FactHttpProbeResult;
import edu.northwestern.dasu.drools.FactNdtProbeResult;
import edu.northwestern.dasu.drools.FactScrapeProbeResult;
import edu.northwestern.dasu.drools.FactSpeedTestResult;
import edu.northwestern.dasu.drools.FactTraceResult;
import edu.northwestern.dasu.drools.FactUpnpTestResult;
import edu.northwestern.dasu.measurement.TestType;
import edu.northwestern.dasu.measurement.DasuTraceRouteModuleV2.TraceEntry;
import edu.northwestern.dasu.timer.ITimer;
import edu.northwestern.dasu.timer.ITimerEvent;
import edu.northwestern.dasu.timer.ITimerEventPerformer;
import edu.northwestern.dasu.util.HashSetCache;
import edu.northwestern.dasu.util.Pair;
import edu.northwestern.dasu.util.Util;


/**
 * @author David Choffnes &lt;drchoffnes@cs.northwestern.edu&gt;
 * 
 *         The Statistics class manages statistics collected by this plugin.
 */
public class Statistics implements ITimerEventPerformer {
	
	private static Statistics self;
	protected final static HashMap<String, Integer> peerMap = new HashMap<String, Integer>();
	private final static HashMap<String, Integer> peerClientMap = new HashMap<String, Integer>();
	private final static HashMap<String, Integer> routerMap = new HashMap<String, Integer>();
	private final static HashMap<String, Integer> urlMap = new HashMap<String, Integer>();
	private static final int MAX_TRACE_RESULTS = 100;
	private static final int MAX_PING_RESULTS = 100;
	private static final int MAX_HTTP_RESULTS = 10;
	private static final int MAX_TEST_RESULTS = 1;
	private static final int MAX_RSTSTAT_RESULTS = 50;
	private static final int MAX_RPCSTAT_RESULTS = 100;
	private final long UPDATE_INTERVAL;
	private boolean DEBUG = false;

	/**
	 * List of things that can be reported.
	 * 
	 */

	/** set of all edges */
	private static HashSet<String> edges;
	/** list of all ping results */
	private static HashSetCache<PingResultDasu> pingResultsDasu;
	/** list of all http results */
	private static HashSetCache<FactHttpProbeResult> httpResultsDasu;
	/** list of all speedTest results */
	private static HashSetCache<FactSpeedTestResult> speedTestResults;
	/** list of all upnpTestResults results */
	private static HashSetCache<FactUpnpTestResult> upnpTestResults;
	/** list of all data reset events */
	private static HashSetCache<RSTStat> RSTStatResults;
	/** list of all connection establishment events */
	private static HashSetCache<RemotePeerConnectionStat> RpcStatResults;
	/** list of all data transfer events */
	private static LinkedList<DataTransferEvent> dataTransferResults;
	/** list of all peers */
	private static HashSet<String> peers;
	/** list of all trace route results */
	private static HashSetCache<FactTraceResult> traceResults;
	/** list of all scrape results */
	private static HashSetCache<FactScrapeProbeResult> scrapeResults;

	private static HashMap<String, Pair<String, Double>> edgePings = new HashMap<String, Pair<String, Double>>();

	/**
	 * Stuff for the database
	 */
	static private DatabaseReporter db = null;

	/** random object for jitter */
	static private Random r = new Random();

	/** maps of internal ids to DB ids */
	static private HashMap<Integer, Integer> peerIds;
	static private HashMap<Integer, Integer> peerClientIds;
	static private HashMap<Integer, Integer> routerIds;
	static private HashMap<Integer, Integer> urlIds;
	private HashSetCache<Exception> errors;
	private int selfId = -1;
	private ITimer timer;
	private Boolean isDumpingStats = false;

	
	private Statistics() {
		if (r == null) {
			r = new Random();
		}
		edges = new HashSet<String>();
		peers = new HashSet<String>();
		traceResults = new HashSetCache<FactTraceResult>(MAX_TRACE_RESULTS);
		pingResultsDasu = new HashSetCache<PingResultDasu>(MAX_PING_RESULTS);
		scrapeResults = new HashSetCache<FactScrapeProbeResult>(MAX_TRACE_RESULTS);
		httpResultsDasu = new HashSetCache<FactHttpProbeResult>(
				MAX_HTTP_RESULTS);
		speedTestResults = new HashSetCache<FactSpeedTestResult>(
				MAX_TEST_RESULTS);
		upnpTestResults = new HashSetCache<FactUpnpTestResult>(MAX_TEST_RESULTS);
		RSTStatResults = new HashSetCache<RSTStat>(MAX_RSTSTAT_RESULTS);
		RpcStatResults = new HashSetCache<RemotePeerConnectionStat>(MAX_RPCSTAT_RESULTS);
		peerIds = new HashMap<Integer, Integer>();
		routerIds = new HashMap<Integer, Integer>();
		peerClientIds = new HashMap<Integer, Integer>();
		urlIds = new HashMap<Integer, Integer>();

		errors = new HashSetCache<Exception>();
		UPDATE_INTERVAL = DasuConfiguration.getInstance()
				.getDbUpdateFreqSec() * 1000;

		timer = MainGeneric.createTimer("OnoStatistics");
		timer.addPeriodicEvent(UPDATE_INTERVAL, this);
		dataTransferResults = new LinkedList<DataTransferEvent>();
	}

	public static synchronized Statistics getInstance() {
		if (self == null) {
			self = new Statistics();
		}

		return self;
	}

	private boolean addPeerClient(String peerClient) {
		if (db == null) {
			db = MainGeneric.getReporter();
			if (!MainGeneric.getReporter().canConnect())
				return false;
		}

		int p;
		
		synchronized (peerClientMap) {
			if (peerClientMap.get(peerClient) == null) {
				peerClientMap.put(peerClient, peerClientMap.size());
			}

			p = peerClientMap.get(peerClient);

			if (!peerClientIds.containsKey(p)) {
				MainGeneric.getReporter().connect();
				int peerClientId = MainGeneric.getReporter().makeId(
						"peerclients", "name", peerClient);

				if (peerClientId == -1) {
					return false;
				}
				MainGeneric.getReporter().disconnect();
				peerClientIds.put(p, peerClientId);
			}
		}

		return true;
	}

	protected boolean addPeer(String peer) {
		if (db == null) {
			db = MainGeneric.getReporter();
			if (!MainGeneric.getReporter().canConnect())
				return false;
		}

		int p;
		
		synchronized (peerMap) {
			if (peerMap.get(peer) == null) {
				peerMap.put(peer, peerMap.size());
			}

			p = peerMap.get(peer);

			if (!peerIds.containsKey(p)) {
				MainGeneric.getReporter().connect();
				int peerId = MainGeneric.getReporter().makeId("peerips", "ip",
						peer);

				if (peerId == -1) {
					return false;
				}
				MainGeneric.getReporter().disconnect();
				peerIds.put(p, peerId);
			}
		}

		return peers.add(peer);
	}
	
	private boolean addUrl(String url) {
		if (db == null) {
			db = MainGeneric.getReporter();
			if (!MainGeneric.getReporter().canConnect())
				return false;
		}

		int p;

		synchronized (urlMap) {
			if (urlMap.get(url) == null) {
				urlMap.put(url, urlMap.size());
			}

			p = urlMap.get(url);
			
			if (!urlIds.containsKey(p)) {
				MainGeneric.getReporter().connect();
				int urlId = MainGeneric.getReporter().makeId("urlids", "url",
						url);

				
				if (urlId == -1) {
					return false;
				}
				MainGeneric.getReporter().disconnect();
				urlIds.put(p, urlId);
			}
		}
		
		return true;
		
	}

	/**
	 * @param map
	 * @param edgeServer
	 * @return
	 */
	private String getString(HashMap<String, Integer> map, int edgeServer) {
		synchronized (map) {
			for (Map.Entry<String, Integer> e : map.entrySet()) {
				if (e.getValue().equals(edgeServer)) {
					return e.getKey();
				}
			}
		}

		return null;
	}


	/**
	 * @return the number of edges seen by this nodes
	 */
	public int getNumEdges() {
		return edges.size();
	}

	/**
	 * @return the list of edges
	 */
	public HashSet<String> getEdgeSet() {
		return edges;
	}

	/**
	 * Called right before plugin unloads.
	 */
	public void flush() {
		if (DasuConfiguration.getInstance().isRecordStats()
				&& DasuConfiguration.getInstance().isSendToDatabase()) {

			MainGeneric.getReporter().setExtendedConnection(true);
			
			batchProcessIds();
			
			commitPingsDasu();
			commitHttpDasu();
			commitDataTransfers();
			commitUpnpTest();
			commitSpeedTest();
			commitRSTStat();
			commitRemotePeerConnections();
			reportTraces();
			reportScrapes();
			commitErrors();
			MainGeneric.getReporter().setExtendedConnection(false);

		}
		MainGeneric.getReporter().disconnect();
	}

	public void addPingResultDasu(String sourceIp, String destIp, Double rtt,
			long timeInMillis, int destType, int taskId) {

		PingResultDasu pr = new PingResultDasu(sourceIp, destIp, rtt,
				timeInMillis, destType, taskId);

		if (DEBUG) {
			System.out.println("AddPingResultDasu " + "destIp: " + destIp
					+ " sourceIp: " + sourceIp + " rtt: " + rtt + " time: "
					+ timeInMillis + ";" + " destType: " + destType + ";"
					+ "taskId: " + taskId);
		}

		synchronized (pingResultsDasu) {
			pingResultsDasu.add(pr);
		}
		synchronized (pingResultsDasu) {
			if (pingResultsDasu.size() >= MAX_PING_RESULTS) {
				commitPingsDasu();
			}
		}

	}

	public void addHttpResultDasu(FactHttpProbeResult httpResult) {

		if (DEBUG) {
			System.out
					.println("AddHttpResultDasu " + "url: " + httpResult.url
							+ " httpResult.downloadDuration: "
							+ httpResult.downloadDuration
							+ " httpResult.time: " + httpResult.time + ";"
							+ " downloadSpeed: " + httpResult.downloadSpeed
							+ ";" + "noBytes: " + httpResult.noBytes + ";"
							+ "taskId: " + httpResult.taskId);
		}
		
		addUrl(httpResult.url);

		synchronized (httpResultsDasu) {
			httpResultsDasu.add(httpResult);
		}
		synchronized (httpResultsDasu) {
			if (httpResultsDasu.size() >= MAX_HTTP_RESULTS) {
				commitHttpDasu();
			}
		}

	}
	
	public void addNdtResultDasu(FactNdtProbeResult ndtResult) {
		commitNdtDasu(ndtResult);
	}

	public void addSpeedTestResult(FactSpeedTestResult result) {

		if (DEBUG) {

			System.out.println("AddSpeedTestResult " + " dl: "
					+ result.downloadSpeed + " " + " ul: " + result.uploadSpeed
					+ " " + " time: " + new Date(result.testTime) + " "
					+ " success: " + result.success + "************");
		}

		synchronized (speedTestResults) {
			speedTestResults.add(result);

			if (speedTestResults.size() >= MAX_TEST_RESULTS)
				commitSpeedTest();
			
		}

	}

	public void addUpnpTestResult(FactUpnpTestResult result) {

		if (DEBUG) {

			System.out.println("UpnpTestResult " + " port: " + result.port
					+ " " + " protocol: " + result.protocol + " " + " time: "
					+ new Date(result.testTime) + " " + " success: "
					+ result.success + "************");
		}

		synchronized (upnpTestResults) {
			upnpTestResults.add(result);
		}
		synchronized (upnpTestResults) {
			if (upnpTestResults.size() >= MAX_TEST_RESULTS) {
				commitUpnpTest();
			}
		}

	}

	public void addRSTStatResult(RSTStat result) {

		if (DEBUG) {

			System.out.println("addConnectionReset " + " did: "
					+ result.downloadId + " " + " numRst: " + result.numRst
					+ " " + " time: " + new Date(result.timeRecorded) + " "
					+ " PeerIp: " + result.peerIp + "************");
		}

		synchronized (Statistics.peerMap) {
			if (!Statistics.peerMap.containsKey(result.peerIp)) {
				Statistics.peerMap.put(result.peerIp, peerMap.size() + 1);
			}
		}

		synchronized (RSTStatResults) {
			RSTStatResults.add(result);
		}
		synchronized (RSTStatResults) {
			if (RSTStatResults.size() >= MAX_RSTSTAT_RESULTS) {
				commitRSTStat();
			}
		}

	}
	
	public void addRpcStatResult(RemotePeerConnectionStat result) {

		if (DEBUG) {

			System.out.println("addRemoteConnection " + " did: "
					+ " " + " time: " + new Date(result.timeRecorded) + " "
					+ " PeerIp: " + result.peerIp + "************");
		}

		synchronized (Statistics.peerMap) {
			if (!Statistics.peerMap.containsKey(result.peerIp)) {
				Statistics.peerMap.put(result.peerIp, peerMap.size() + 1);
			}
		}

		synchronized (RpcStatResults) {
			RpcStatResults.add(result);
		}
		synchronized (RpcStatResults) {
			if (RpcStatResults.size() >= MAX_RPCSTAT_RESULTS) {
				commitRemotePeerConnections();
			}
		}

	}
	
	public void addScrapeResult(FactScrapeProbeResult result) {

		if (DEBUG) {

			System.out.println("addScrapeResult " + " type " 
					+ result.type + " time: " + new Date(result.time) + " "
					+ " numLeechers: " + result.getNumLeechers()
					+ " numSeeders: "  + result.getNumSeeds() + "************");
		}
		
		
		synchronized (scrapeResults) {
			if (scrapeResults.size() > MAX_TRACE_RESULTS) {
				reportScrapes();
			}
			scrapeResults.add(result);
		}

	}
	
	public void reportScrapes() {

		if (DEBUG)
			System.out.println("reportScrapesDasu()");
		
		if (db == null) {
			db = MainGeneric.getReporter();
		}
		if (!MainGeneric.getReporter().connect()
				|| !DasuConfiguration.getInstance().isSendToDatabase())
			return;
		
		while (scrapeResults.size() > 0) {
			FactScrapeProbeResult scrape = null;
			// update map of ids from database
			synchronized (scrapeResults) {
				Iterator<FactScrapeProbeResult> it = scrapeResults.iterator();
				scrape = it.next();
				it.remove();
			}
		
			for (String ip : scrape.getIps())
				addPeerForLookup(ip);
			
			//batchProcessIds();
			int myId = updateIdMap();
			if (myId == -1)
					return;	
	
			synchronized (MainGeneric.getReporter()) {
				// make top-level entry
				long entryId = MainGeneric.getReporter().insertScrapeEntry(
						scrape.type, myId, scrape.tracker, scrape.hash, scrape.getIps().size(), 
						scrape.numSeeds, scrape.numLeechers, scrape.taskId, scrape.time);
	
				// if valid, insert scrape entries
				if (entryId > -1) {
					for (int i=0; i < scrape.getIps().size(); i++) {
			
						String ip = scrape.getIps().get(i);
						int peerid = -1;
			
						// get mappings for storing entry
						addPeer(ip);
						int tempPeerIp = peerMap.get(ip);
						if (!peerIds.containsKey(tempPeerIp)) {
							peerid = MainGeneric.getReporter().makeId("peerips", "ip",
									getString(peerMap, tempPeerIp));
			
							if (peerid == -1) {
								continue;
							}
			
							peerIds.put(tempPeerIp, peerid);
						} else
							peerid = peerIds.get(tempPeerIp);
					
						// insert data
						MainGeneric.getReporter().insertScrapeEntryData(entryId,
								peerid, scrape.getPorts().get(i), scrape.getUdpPorts().get(i));
							
					}
				} // end case valid entry
				else {
					scrapeResults.add(scrape);
					break; // exit for now
				}
				
			}// end synchronized part
		}
		
		MainGeneric.getReporter().disconnect();

	}

	private void commitDataTransfers() {
		int sourceId = -1;
		int destId = -1;
		int currentId = -1;
		synchronized (isDumpingStats) {

			isDumpingStats = true;
			if (db == null) {
				db = MainGeneric.getReporter();
			}
			if (!MainGeneric.getReporter().connect())
				return;

			LinkedList<DataTransferEvent> dtesToRemove = new LinkedList<DataTransferEvent>();
			LinkedList<DataTransferEvent> localCopy = new LinkedList<DataTransferEvent>();
			for (int i = 0; i < dataTransferResults.size(); i++)
				localCopy.add(dataTransferResults.get(i));

			batchProcessIds();
			
			// update map of ids from database
			// synchronized (dataTransferResults) {
			for (DataTransferEvent dte : localCopy) {
				if (!peerIds.containsKey(dte.source)) {
					System.out.println("commitDataTransfers(): makeId()");
					sourceId = MainGeneric.getReporter().makeId("peerips",
							"ip", getString(peerMap, dte.source));

					if (sourceId == -1) {
						continue;
					}

					peerIds.put(dte.source, sourceId);
				} else {
					sourceId = peerIds.get(dte.source);
				}

				if (!peerIds.containsKey(dte.dest)) {
					destId = MainGeneric.getReporter().makeId("peerips", "ip",
							getString(peerMap, dte.dest));

					if (destId == -1) {
						continue;
					}

					peerIds.put(dte.dest, destId);
				} else {
					destId = peerIds.get(dte.dest);
				}

				if (!peerIds.containsKey(dte.current)) {
					currentId = MainGeneric.getReporter().makeId("peerips",
							"ip", getString(peerMap, dte.current));

					if (currentId == -1) {
						continue;
					}

					peerIds.put(dte.current, currentId);
				} else {
					currentId = peerIds.get(dte.current);
				}

				try {
					if (MainGeneric.getReporter().insertDataTransferResult(
							dte.time, dte.experimentId, dte.type, dte.size,
							sourceId, destId, currentId, dte.raceResultPrimary,
							dte.raceResultSecondary)) {
						dtesToRemove.add(dte);
					}
				} catch (Exception e) {
					e.printStackTrace();
					dtesToRemove.add(dte);
				}
			}
			// }

			for (DataTransferEvent dte : dtesToRemove) {
				synchronized (dataTransferResults) {
					dataTransferResults.remove(dte);
				}
			}

			MainGeneric.getReporter().disconnect();
			isDumpingStats = false;
		}
	}

	/**
    *
    */
	private void commitPingsDasu() {

		if (DEBUG)
			System.out.println("commitPingsDasu()");

		if (db == null) {
			db = MainGeneric.getReporter();
			if (!MainGeneric.getReporter().canConnect()
					|| !DasuConfiguration.getInstance().isSendToDatabase())
				return;
		}

		LinkedList<PingResultDasu> pingsToRemove = new LinkedList<PingResultDasu>();
		// update map of ids from database
		synchronized (pingResultsDasu) {
			
			for (PingResultDasu p : pingResultsDasu) {
				if (db == null) {
					db = MainGeneric.getReporter();
				}
				
				if (!peerIds.containsKey(peerMap.get(p.sourceIp))) {
					
					if (DEBUG)
						System.out.println("commitPingDasu: makeId()");
					
					p.sourceId = MainGeneric.getReporter().makeId("peerips",
							"ip", p.sourceIp);

					if (p.sourceId == -1)
						return;

					peerIds.put(peerMap.get(p.sourceIp), p.sourceId);
				}

				p.sourceId = peerIds.get(peerMap.get(p.sourceIp));

				if (p.destType == 1)
					p.destId = peerIds.get(peerMap.get(p.destIp));
				
				if (p.destId != -1 && p.sourceId != -1)
					if (MainGeneric.getReporter().insertPingResultDasu(
							p.sourceId, p.destId, p.rtt, p.time, p.destType,
							p.taskId))
						pingsToRemove.add(p);

			}

		}

		synchronized (pingResultsDasu) {
			pingResultsDasu.removeAll(pingsToRemove);
		}
		MainGeneric.getReporter().disconnect();

	}

	public int makeUUIDid(String uuid) {
		return MainGeneric.getReporter().makeId("uuids", "uuid", uuid);
	}
	
	/**
    *
    */
	private void commitHttpDasu() {

		if (DEBUG)
			System.out.println("commitHttpDasu()");

		if (db == null) {
			db = MainGeneric.getReporter();
			if (!MainGeneric.getReporter().canConnect()
					|| !DasuConfiguration.getInstance().isSendToDatabase())
				return;
		}
		
		LinkedList<FactHttpProbeResult> httpsToRemove = new LinkedList<FactHttpProbeResult>();
		
		int myId = updateIdMap();
		if (myId == -1)
			return;
		
		synchronized (httpResultsDasu) {
			for (FactHttpProbeResult h : httpResultsDasu) {
				if (db == null)
					db = MainGeneric.getReporter();
				
				int urlId = -1;
				if (!urlIds.containsKey(urlMap.get(h.url))) {
					System.out.println("commitHttpDasu: makeId()");
					urlId = MainGeneric.getReporter().makeId("urlids", "url", h.url);
					
					if (urlId == -1)
						continue;
					
					urlIds.put(urlMap.get(h.url), urlId);
				} else
					urlId = urlIds.get(urlMap.get(h.url));

				if (MainGeneric.getReporter().insertHttpResultDasu(myId,
						urlId, h.downloadDuration, h.downloadSpeed, h.time,
						h.noBytes, h.taskId))
					httpsToRemove.add(h);

			}
		}

		synchronized (httpResultsDasu) {
			httpResultsDasu.removeAll(httpsToRemove);
		}
		MainGeneric.getReporter().disconnect();

	}

	/**
    *
    */
	public void commitNdtDasu(FactNdtProbeResult ndt) {

		if (DEBUG)
			System.out.println("commitNdtDasu()");

		if (db == null) {
			db = MainGeneric.getReporter();
			if (!MainGeneric.getReporter().canConnect()
					|| !DasuConfiguration.getInstance().isSendToDatabase())
				return;
		}

		int myId = updateIdMap();
		if (myId == -1)
			return;
		
//		if (ndt.failed == true)
//			return;

		MainGeneric.getReporter().insertNdtResultDasu(myId,
				ndt.osName, ndt.osArch, ndt.osVer, ndt.interfaceName, ndt.javaVer, 
				ndt.server, ndt.c2sRate, ndt.s2cRate, ndt.mylink, ndt.mismatch, ndt.avgrtt, 
				ndt.packetSize, ndt.connectionStalledTimes, ndt.connectionIdlePercentage, 
				boolToInt(ndt.c2sPacketQueuing), boolToInt(ndt.s2cPacketQueuing), 
				ndt.s2cThroughputPacketQueueingDetectedPercentage, 
				ndt.connectionReceiverLimitedPercentage, ndt.connectionSenderLimitedPercentage,
				ndt.connectionNetworkLimitedPercentage, boolToInt(ndt.RFC2018), boolToInt(ndt.RFC896), 
				boolToInt(ndt.RFC3168), boolToInt(ndt.RFC1323TimeStamping), boolToInt(ndt.RFC1323WindowScaling), 
				ndt.serverScalingFactor, ndt.clientScalingFactor, boolToInt(ndt.serverBehindFirewall), 
				boolToInt(ndt.clientBehindFirewall), ndt.theoreticalNetworkLimit, ndt.serverBuffer, 
				ndt.serverThroughputLimit, ndt.clientBuffer, ndt.clientThroughputLimit, 
				ndt.time, boolToInt(ndt.failed), ndt.errorMsg, ndt.c2sRateNetstat, ndt.s2cRateNetstat, 
				ndt.networkFlowControlThroughputLimit, boolToInt(ndt.congestion), ndt.congestionSignals, ndt.fastRetransmit, ndt.timeouts);

		MainGeneric.getReporter().disconnect();

	}
	
	
	/**
    *
    */
	public void commitSpeedTest() {

		if (DEBUG)
			System.out.println("commitSpeedTest()");

		if (db == null) {
			db = MainGeneric.getReporter();
			if (!MainGeneric.getReporter().canConnect()
					|| !DasuConfiguration.getInstance().isSendToDatabase())
				return;
		}

		LinkedList<FactSpeedTestResult> speedTestsToRemove = new LinkedList<FactSpeedTestResult>();
		
		int myId = updateIdMap();
		if (myId == -1)
			return;

		synchronized (speedTestResults) {
			for (FactSpeedTestResult s : speedTestResults) {
				if (db == null)
					db = MainGeneric.getReporter();

				if (!s.isSuccess()) {
					speedTestsToRemove.add(s);
					continue;
				}
				
				if (MainGeneric.getReporter().insertSpeedTestResult(myId,
						s.testTime, s.downloadSpeed, s.uploadSpeed,
						boolToInt(s.encryption), testTypeToString(s.testType), 
						s.downloadSpeedNetstat, s.uploadSpeedNetstat))
					speedTestsToRemove.add(s);

			}
		}

		synchronized (speedTestResults) {
			speedTestResults.removeAll(speedTestsToRemove);
		}
		MainGeneric.getReporter().disconnect();

	}

	private char testTypeToString(TestType testType) {
		char ret = 'u';
		if ((testType.compareTo(TestType.AZUREUS_DL) == 0)
				|| (testType.compareTo(TestType.AZUREUS_DL_ENCRYPTION)) == 0)
			ret = 'd';

		return ret;
	}


	public int updateIdMap() {

		batchProcessIds();
		String myIp = Main.getPublicIpAddress();
		int myId = -1;

		if (!peerIds.containsKey(peerMap.get(myIp))) {
			System.out.println("updateIdMap: makeId()");
			myId = MainGeneric.getReporter().makeId("peerips", "ip", myIp);

			if (myId != -1)
				peerIds.put(peerMap.get(myIp), myId);
			
		} else
			myId = peerIds.get(peerMap.get(myIp));
		
		return myId;
	}
	/**
    *
    */
	public void commitUpnpTest() {

		if (DEBUG)
			System.out.println("commitUpnpTest()");

		if (db == null) {
			db = MainGeneric.getReporter();
			if (!MainGeneric.getReporter().canConnect()
					|| !DasuConfiguration.getInstance().isSendToDatabase())
				return;
		}

		LinkedList<FactUpnpTestResult> upnpTestsToRemove = new LinkedList<FactUpnpTestResult>();

//		batchProcessIds();		
//		String myIp = Main.getPublicIpAddress();
//		int myId = -1;
//		// update map of ids from database
//		if (!peerIds.containsKey(peerMap.get(myIp))) {
//			myId = MainGeneric.getReporter().makeId("peerips", "ip", myIp);
//
//			if (myId == -1)
//				return;
//
//			peerIds.put(peerMap.get(myIp), myId);
//		} else
//			myId = peerIds.get(peerMap.get(myIp));
		
		int myId = updateIdMap();
		if (myId == -1)
			return;

		synchronized (upnpTestResults) {
			for (FactUpnpTestResult u : upnpTestResults) {
				if (db == null)
					db = MainGeneric.getReporter();

				if (MainGeneric.getReporter().insertUpnpTestResult(myId,
						u.testTime, u.port, u.protocol, boolToInt(u.success)))
					upnpTestsToRemove.add(u);

			}
		}

		synchronized (upnpTestResults) {
			upnpTestResults.removeAll(upnpTestsToRemove);
		}
		MainGeneric.getReporter().disconnect();

	}

	private int boolToInt(boolean b) {
		if (b)
			return 1;

		return 0;
	}

	/**
    *
    */
	private void commitRSTStat() {

		if (DEBUG)
			System.out.println("commitRSTStat()");

		if (db == null) {
			db = MainGeneric.getReporter();
			if (!MainGeneric.getReporter().canConnect()
					|| !DasuConfiguration.getInstance().isSendToDatabase())
				return;
		}

		LinkedList<RSTStat> RSTStatsToRemove = new LinkedList<RSTStat>();
		
		int myId = updateIdMap();
		if (myId == -1)
			return;

		synchronized (RSTStatResults) {
			for (RSTStat r : RSTStatResults) {
				if (db == null)
					db = MainGeneric.getReporter();

				int peerId = -1;

				if (!peerIds.containsKey(peerMap.get(r.peerIp))) {
					System.out.println("commitRSTStat: makeId()");
					peerId = MainGeneric.getReporter().makeId("peerips", "ip",
							r.peerIp);
					if (peerId == -1)
						return;

					peerIds.put(peerMap.get(r.peerIp), peerId);
				}

				peerId = peerIds.get(peerMap.get(r.peerIp));

				if (MainGeneric.getReporter().insertRSTStat(peerId, myId, r.downloadId, r.numRst, r.timeRecorded,
					r.type, r.port, r.bitTorrentClient, r.source, boolToInt(r.isSeed), boolToInt(r.amSeed), 
					r.numSeeds, r.numPeers, r.encryption))
					RSTStatsToRemove.add(r);

			}
		}

		synchronized (RSTStatResults) {
			RSTStatResults.removeAll(RSTStatsToRemove);
		}
		MainGeneric.getReporter().disconnect();

	}
	
	/**
    *
    */
	private void commitRemotePeerConnections() {

		if (DEBUG)
			System.out.println("commitRemotePeerConnections()");

		if (db == null) {
			db = MainGeneric.getReporter();
			if (!MainGeneric.getReporter().canConnect()
					|| !DasuConfiguration.getInstance().isSendToDatabase())
				return;
		}


		LinkedList<RemotePeerConnectionStat> RemotePeerRemove = new LinkedList<RemotePeerConnectionStat>();
		
		int myId = updateIdMap();
		if (myId == -1)
			return;

		synchronized (RpcStatResults) {
			for (RemotePeerConnectionStat r : RpcStatResults) {
				if (db == null)
					db = MainGeneric.getReporter();

				int peerId = -1;

				if (!peerIds.containsKey(peerMap.get(r.peerIp))) {
					System.out.println("commitRemotePeer: makeId()");
					peerId = MainGeneric.getReporter().makeId("peerips", "ip",
							r.peerIp);
					if (peerId == -1)
						return;

					peerIds.put(peerMap.get(r.peerIp), peerId);
				}

				peerId = peerIds.get(peerMap.get(r.peerIp));

				if (MainGeneric.getReporter().insertRemotePeerConnectionStat(peerId, myId, r.timeRecorded,
					r.type, r.port, r.bitTorrentClient, boolToInt(r.isSeed), boolToInt(r.amSeed), 
					r.numSeeds, r.numPeers, r.encryption))
					RemotePeerRemove.add(r);

			}
		}

		synchronized (RpcStatResults) {
			RpcStatResults.removeAll(RemotePeerRemove);
		}
		MainGeneric.getReporter().disconnect();

	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * org.gudy.azureus2.plugins.utils.UTTimerEventPerformer#perform(org.gudy
	 * .azureus2.plugins.utils.UTTimerEvent)
	 */
	public void perform(ITimerEvent event) {
		try {
			if (DasuConfiguration.getInstance().isRecordStats()
					&& DasuConfiguration.getInstance().isSendToDatabase()) {
				MainGeneric.getReporter().setExtendedConnection(true);

				reportTraces();
				traceResults.clear();
				reportScrapes();
				scrapeResults.clear();
				commitErrors();
				errors.clear();
				commitPingsDasu();
				pingResultsDasu.clear();
				commitHttpDasu();
				httpResultsDasu.clear();
				commitSpeedTest();
				speedTestResults.clear();
				commitUpnpTest();
				upnpTestResults.clear();
				commitRSTStat();
				RSTStatResults.clear();
				commitRemotePeerConnections();
				RpcStatResults.clear();

				MainGeneric.getReporter().setExtendedConnection(false);

			} else {
				errors.clear();
				traceResults.clear();
				pingResultsDasu.clear();
			}
		} catch (Exception e) {
			e.printStackTrace();
		}

		MainGeneric.getReporter().disconnect();
	}

	private void commitErrors() {
		if (db == null) {
			db = MainGeneric.getReporter();
		}
		if (!MainGeneric.getReporter().connect())
			return;
		while (errors.size() > 0) {
			Exception e = errors.iterator().next();
			MainGeneric.getReporter().reportError(e);
			errors.remove(e);
		}

		MainGeneric.getReporter().disconnect();

	}

	public void addTraceRouteResult(FactTraceResult tr) {

		addPeerForLookup(tr.dest);
		
		for (int i=0; i< tr.entries.size(); i++) {
			TraceEntry te = tr.entries.get(i);
			
			for (int j = 0; j < te.router.length; j++) {
				String router;

				if (te.numRouters - 1 < j) {
					router = te.router[0];
				} else {
					router = te.router[j];
				}

				addRouterForLookup(router);
			}
		}
		
		synchronized (traceResults) {
			if (traceResults.size() > MAX_TRACE_RESULTS) {
				reportTraces();
			}
			traceResults.add(tr);
		}
		
	}
	
	private void reportTraces() {
		int sourceId = -1;
		int destId = -1;

		if (db == null) {
			db = MainGeneric.getReporter();
		}
		if (!MainGeneric.getReporter().connect()
				|| !DasuConfiguration.getInstance().isSendToDatabase())
			return;

		batchProcessIds();
		
		while (traceResults.size() > 0) {
			FactTraceResult tr = null;
			// update map of ids from database
			synchronized (traceResults) {
				Iterator<FactTraceResult> it = traceResults.iterator();
				tr = it.next();
				it.remove();
			}

			// get mappings for storing entry
			addPeer(tr.source);
			int tempPeerIp = peerMap.get(tr.source);
			if (!peerIds.containsKey(tempPeerIp)) {
				sourceId = MainGeneric.getReporter().makeId("peerips", "ip",
						getString(peerMap, tempPeerIp));

				if (sourceId == -1) {
					continue;
				}

				peerIds.put(tempPeerIp, sourceId);
			} else {
				sourceId = peerIds.get(tempPeerIp);
			}

			addPeer(tr.dest);
			tempPeerIp = peerMap.get(tr.dest);
			if (!peerIds.containsKey(tempPeerIp)) {
				destId = MainGeneric.getReporter().makeId("peerips", "ip",
						getString(peerMap, tempPeerIp));

				if (destId == -1) {
					continue;
				}

				peerIds.put(tempPeerIp, destId);
			} else {
				destId = peerIds.get(tempPeerIp);
			}

			int lastIndex = tr.entries.size() - 1;
			float rtts[];
			for (int i = tr.entries.size() - 1; i >= 0; i--) {
				rtts = tr.entries.get(i).rtt;
				if (rtts[0] + rtts[1] + rtts[2] == -3) {
					lastIndex = i;
				} else {
					break;
				}
			}

			int taskId = tr.taskId;

			synchronized (MainGeneric.getReporter()) {
				// make top-level entry
				long entryId = MainGeneric.getReporter().insertTraceEntry(
						sourceId, destId, tr.timestamp, lastIndex, taskId);

				// if valid, insert trace route entries
				if (entryId > -1) {

					TraceEntry te;
					for (int j = 0; j <= lastIndex; j++) {
						te = tr.entries.get(j);
						int myRouterIds[] = new int[3];
						for (int i = 0; i < te.router.length; i++) {
							String router;

							if (te.numRouters - 1 < i) {
								router = te.router[0];
							} else {
								router = te.router[i];
							}

							if (!routerMap.containsKey(router)) {
								routerMap.put(router, routerMap.size());
							}
							int tempRouterId = routerMap.get(router);
							if (routerIds.get(tempRouterId) == null) {
								myRouterIds[i] = MainGeneric.getReporter()
										.makeId(
												"routerips",
												"ip",
												getString(routerMap,
														tempRouterId));

								if (myRouterIds[i] == -1) {
									continue;
								}

								routerIds.put(tempRouterId, myRouterIds[i]);
							} else {
								myRouterIds[i] = routerIds.get(tempRouterId);
							}
						} // end for router

						// insert data
						MainGeneric.getReporter().insertTraceEntryData(entryId,
								myRouterIds, te);
					} // end for trace entries

				} // end case valid entry
				else {
					traceResults.add(tr);
					break; // exit for now
				}
			} // end synchronized part

		}

		MainGeneric.getReporter().disconnect();

	}

	protected int getPeerId(String ip) {
		if (!DasuConfiguration.getInstance().isRecordStats()
				|| !DasuConfiguration.getInstance().isSendToDatabase())
			return -1;
		
		if (peerMap.get(ip) == null
				|| (peerIds.get(peerMap.get(ip)) == null || peerIds.get(peerMap
						.get(ip)) == -1)) {
			if (!addPeer(ip))
				return -1;
		}
		if (peerIds.get(peerMap.get(ip)) == null) {
			return -1;
		}
		return peerIds.get(peerMap.get(ip));
	}

	protected int getPeerClientId(String client) {
		if (!DasuConfiguration.getInstance().isRecordStats()
				|| !DasuConfiguration.getInstance().isSendToDatabase())
			return -1;
		if (client.matches("(.*)\\[(.*)\\]") || client.contains("Unknown")
				|| client.contains("Inconnu") || client.contains("Ukjent")
				|| client.contains("Unbekannt")
				|| client.contains("Sconosciuto")) {
			if (client.toLowerCase().contains("exeem"))
				client = "Exeem";
			else
				client = "Unknown";
		}

		if (peerClientMap.get(client) == null) {
			if (!addPeerClient(client))
				return -1;
		}
		if (peerClientIds.get(peerClientMap.get(client)) == null)
			return -1;
		else
			return peerClientIds.get(peerClientMap.get(client));
	}

	public Map<String, Pair<String, Double>> getEdgeRtts() {
		return edgePings;
	}

	public int getSelfId() {
		// if (selfId == -1){
		selfId = getPeerId(MainGeneric.getPublicIpAddress());
		// }
		return selfId;
	}

	public void stop() {
		MainGeneric.getReporter().flush();
		if (timer != null) {
			timer.destroy();
		}
		self = null;
		synchronized (dataTransferResults) {
			dataTransferResults.clear();
		}
		db = null;
		edges.clear();
		traceResults.clear();
		scrapeResults.clear();

		peerClientIds.clear();
		peerClientMap.clear();
		peerIds.clear();
		peerMap.clear();
		peers.clear();

		r = null;
		routerIds.clear();
		routerMap.clear();
		urlIds.clear();
		urlMap.clear();

	}

	private int batchProcessTable(String tableName, String columnName,
			HashMap<String, Integer> localMap, HashMap<Integer, Integer> dbMap) {

		int numRemaining = 0;
		synchronized (localMap) {
			LinkedList<String> keys = new LinkedList<String>();
			for (Entry<String, Integer> ent : localMap.entrySet()) {
				if (!dbMap.containsKey(ent.getValue())) {
					keys.add(ent.getKey());
				}
			}

			if (keys.size() == 0)
				return 0;
			Map<String, Integer> dbKeys = MainGeneric.getReporter()
					.makeIdBatch(tableName, columnName, keys);

			if (dbKeys == null)
				return 0;

			int numFound = 0;
			for (Entry<String, Integer> ent : dbKeys.entrySet()) {
				dbMap.put(localMap.get(ent.getKey()), ent.getValue());
				numFound++;
			}
			numRemaining = keys.size() - numFound;
		}
		return numRemaining;
	}

	protected void batchProcessIds() {
		if (!DasuConfiguration.getInstance().isRecordStats()
				|| !DasuConfiguration.getInstance().isSendToDatabase())
			return;

		int totalLeft;
		int numLoops = 0; // to ensure termination
		do {
			totalLeft = 0;
			totalLeft += batchProcessTable("peerips", "ip", peerMap, peerIds);
			totalLeft += batchProcessTable("peerclients", "name",
					peerClientMap, peerClientIds);
			totalLeft += batchProcessTable("routerips", "ip", routerMap,
					routerIds);
			totalLeft += batchProcessTable("urlids", "url", urlMap, urlIds);
						
			numLoops++;
		} while (totalLeft > 0 && MainGeneric.getReporter().canConnect()
				&& numLoops < 1000);

	}

	/**
	 * Add peer only for lookup, don't connect to DB to get id
	 * 
	 * @param peer
	 * @return
	 */
	public void addPeerForLookup(String peer) {

		synchronized (peerMap) {
			if (peerMap.get(peer) == null) {
				peerMap.put(peer, peerMap.size());
			}

		}
		peers.add(peer);
	}

	/**
	 * Register peer client name for DB lookup
	 * 
	 * @param peerClient
	 */
	protected void addPeerClientForLookup(String peerClient) {
		synchronized (peerClientMap) {
			if (peerClientMap.get(peerClient) == null) {
				peerClientMap.put(peerClient, peerClientMap.size());
			}
		}

	}

	/**
	 * Register peer client name for DB lookup
	 * 
	 * @param peerClient
	 */
	public void addRouterForLookup(String routerIp) {
		synchronized (routerMap) {
			if (routerMap.get(routerIp) == null) {
				routerMap.put(routerIp, routerMap.size());
			}
		}

	}

	public void reportIpChange(String oldIp, String newIp, String localIp,
			long currentGMTTime) {
		int pub = getPeerId(oldIp);
		int nouvel = getPeerId(newIp);
		int local = getPeerId(localIp);
		MainGeneric.getReporter().reportIpChange(pub, nouvel, local,
				currentGMTTime);

	}
	
	/**
    *
    */
	public void commitUserComments(String dids, String comment) {

		if (DEBUG)
			System.out.println("commitUserComments()");

		if (db == null) {
			db = MainGeneric.getReporter();
			if (!MainGeneric.getReporter().canConnect()
					|| !DasuConfiguration.getInstance().isSendToDatabase())
				return;
		}

		int myId = updateIdMap();
		if (myId == -1)
			return;

		MainGeneric.getReporter().insertUserReports(myId, dids, comment);

		MainGeneric.getReporter().disconnect();

	}
	
	/**
    *
    */
	public void commitTask(int taskId) {

		if (DEBUG)
			System.out.println("commitTask()");

		if (db == null) {
			db = MainGeneric.getReporter();
			if (!MainGeneric.getReporter().canConnect()
					|| !DasuConfiguration.getInstance().isSendToDatabase())
				return;
		}

		int myId = updateIdMap();
		if (myId == -1)
			return;

		MainGeneric.getReporter().insertTask(myId, Util.currentGMTTime(), taskId);

		MainGeneric.getReporter().disconnect();

	}

	/**
    *
    */
	public void commitIspInfo(String prefix, String ispname) {

		if (DEBUG)
			System.out.println("commitIspInfo()");

		if (db == null) {
			db = MainGeneric.getReporter();
			if (!MainGeneric.getReporter().canConnect()
					|| !DasuConfiguration.getInstance().isSendToDatabase())
				return;
		}

		int myId = updateIdMap();
		if (myId == -1)
			return;

		MainGeneric.getReporter().insertIspInfo(myId, prefix, ispname);

		MainGeneric.getReporter().disconnect();

	}
	
	/**
    *
    */
    public void commitInterference(String type, Long timestamp) {

		if (DEBUG)
			System.out.println("commitInterference()");

		if (db == null) {
			db = MainGeneric.getReporter();
			if (!MainGeneric.getReporter().canConnect()
					|| !DasuConfiguration.getInstance().isSendToDatabase())
				return;
		}

		int myId = updateIdMap();
		if (myId == -1)
			return;

		MainGeneric.getReporter().insertInterference(myId, type, timestamp);

		MainGeneric.getReporter().disconnect();

	}
    
   	/**
    *
    */
	public void commitInterferenceEvent(String type, Long timestamp, Long downloadId, Long droppedTo, float mean) {

		if (DEBUG)
			System.out.println("commitInterferenceEvent()");

		if (db == null) {
			db = MainGeneric.getReporter();
			if (!MainGeneric.getReporter().canConnect()
					|| !DasuConfiguration.getInstance().isSendToDatabase())
				return;
		}

		int myId = updateIdMap();
		if (myId == -1)
			return;

		MainGeneric.getReporter().insertInterferenceEvent(myId, type, timestamp, downloadId, droppedTo, mean);

		MainGeneric.getReporter().disconnect();

	}


}
